!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   structure_check_mod
!> @brief   check protein structure
!! @authors Takaharu Mori (TM)
! 
!  (c) Copyright 2020 RIKEN. All rights reserved.
! 
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module structure_check_mod

  use molecules_str_mod
  use mpi_parallel_mod
  use constants_mod
  use messages_mod
  use string_mod

  implicit none
  private

  ! for ring check
  integer,                   save  :: total_nring_groups
  integer,                   save  :: nring_types
  integer,                   save  :: max_nrings_in_residue
  real(wp),     allocatable, save  :: criteria(:,:)
  integer,      allocatable, save  :: num_rings(:)
  integer,      allocatable, save  :: head(:)
  integer,      allocatable, save  :: NA(:,:,:)
  character(4), allocatable, save  :: RA(:,:,:) 
  character(4), allocatable, save  :: EA(:,:,:)
  character(6), allocatable, save  :: ring_resname(:)
  integer,      allocatable, save  :: ring_list(:)
  integer,      allocatable, save  :: grp2resid_ring(:)
  integer,      allocatable, save  :: grp2resno_ring(:)
  real(wp),     allocatable, save  :: grp2criteria_ring(:)
  integer,                   save  :: nexclusion_ring
  integer,      allocatable, save  :: exclist_ring(:)

  ! for chilarity check
  integer,                   save  :: max_chiral_members
  integer,                   save  :: max_chirals_in_residue
  integer,                   save  :: total_nchiral_groups
  integer,                   save  :: nchiral_types
  integer,      allocatable, save  :: num_chirals(:)
  integer,      allocatable, save  :: nchiral_members(:)
  character(4), allocatable, save  :: CA(:,:,:)
  character(6), allocatable, save  :: chiral_resname(:,:)
  integer,      allocatable, save  :: chiral_list(:)
  character(6), allocatable, save  :: grp2resname_chiral(:)
  integer,      allocatable, save  :: grp2resno_chiral(:)
  integer,                   save  :: nexclusion_chiral
  integer,      allocatable, save  :: exclist_chiral(:)

  ! for cis-bond check
  ! TODO

  ! subroutines
  public   :: setup_structure_check
  private  :: setup_charmm
  private  :: setup_charmm19
  private  :: setup_ring_atomlists
  private  :: setup_chiral_atomlists
  public   :: perform_structure_check
  private  :: check_ring_structure
  private  :: check_chirality

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_structure_check
  !> @brief        setup for structure check
  !! @authors      TM
  !! @param[in]    molecule : molecule information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_structure_check(molecule, forcefield, coord, check, &
                 fix_ring, fix_chiral, exclude_ring_grpid, exclude_chiral_grpid)

    ! formal arguments
    type(s_molecule),   intent(inout) :: molecule
    character(20),      intent(in)    :: forcefield
    real(wp),           intent(inout) :: coord(:,:)
    logical,            intent(in)    :: check
    logical,            intent(in)    :: fix_ring
    logical,            intent(in)    :: fix_chiral
    character(MaxLine), intent(in)    :: exclude_ring_grpid
    character(MaxLine), intent(in)    :: exclude_chiral_grpid

    ! local variables
    logical           :: warning

    return

  end subroutine setup_structure_check

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_charmm
  !! @authors      TM
  !! @param[in]    molecule : molecule information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_charmm

    return

  end subroutine setup_charmm

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_charmm19
  !! @authors      TM
  !! @param[in]    molecule : molecule information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_charmm19

    return

  end subroutine setup_charmm19

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_ring_atomlists
  !! @authors      TM
  !! @param[in]    molecule : molecule information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_ring_atomlists(molecule, exclude_ring_grpid)

    ! formal arguments
    type(s_molecule),     target, intent(in) :: molecule
    character(MaxLine),           intent(in) :: exclude_ring_grpid

    ! local variables
    integer                :: i, j, k, l, m, jres
    integer                :: prev_head, prev_jres, prev_resno, curr_resno
    integer                :: icycle, ifound_grp, ifound_atm
    logical                :: residue_changed, is_ring_residue
    character(4)           :: curr_segid, prev_segid
    integer                :: natoms

    integer,       pointer :: resno(:)
    character(4),  pointer :: segid(:)
    character(4),  pointer :: atomname(:)
    character(6),  pointer :: resname(:)

    return

  end subroutine setup_ring_atomlists

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_chiral_atomlists
  !! @authors      TM
  !! @param[in]    molecule : molecule information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_chiral_atomlists(molecule, exclude_chiral_grpid)

    ! formal arguments
    type(s_molecule), target, intent(in) :: molecule
    character(MaxLine),       intent(in) :: exclude_chiral_grpid

    ! local variables
    integer                :: i, j, k, l, m
    integer                :: icycle, ifound_grp, jtype, lcount
    integer                :: ihead0, prev_ncatoms
    integer                :: prev_jres, prev_resno, curr_resno
    logical                :: residue_changed, is_chiral_residue
    character(4)           :: curr_segid, prev_segid

    integer                :: natoms

    integer,       pointer :: resno(:)
    character(4),  pointer :: segid(:)
    character(4),  pointer :: atomname(:)
    character(6),  pointer :: resname(:)

    return

  end subroutine setup_chiral_atomlists

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    perform_structure_check
  !> @brief        check protein structure
  !! @authors      TM
  !! @param[in]    molecule : molecule information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine perform_structure_check(coord, check, fix_ring, fix_chiral, warning)

    ! formal arguments
    real(wp), intent(inout) :: coord(:,:)
    logical,  intent(in)    :: check
    logical,  intent(in)    :: fix_ring
    logical,  intent(in)    :: fix_chiral
    logical,  intent(in)    :: warning

    return

  end subroutine perform_structure_check

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    check_ring_structure
  !! @authors      TM
  !! @param[in]    molecule : molecule information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine check_ring_structure(coord, check, fix, warning)

    ! formal arguments
    real(wp), intent(inout) :: coord(:,:)
    logical,  intent(in)    :: check
    logical,  intent(in)    :: fix
    logical,  intent(in)    :: warning

    ! local variables
    integer  :: i, j
    integer  :: i1, i2, nr, ne
    integer  :: ihead, icount, jcount
    logical  :: sus_found, def_error, do_fix

    return

  end subroutine check_ring_structure

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    check_chirality
  !> @brief        check chirality in amino acids and RNA/DNA
  !! @authors      TM
  !! @param[in]    coord : coordinates of atoms in the system
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine check_chirality(coord, check, fix, warning)

    ! formal arguments
    real(wp), intent(inout) :: coord(:,:)
    logical,  intent(in)    :: check
    logical,  intent(in)    :: fix
    logical,  intent(in)    :: warning

    ! local variables
    integer  :: i, j, i1, i2, i3, i4, i5, icount
    real(wp) :: vec42(3), vec43(3), vec15(3), nvec(3), tmp
    logical  :: sus_found, do_fix

    return

  end subroutine check_chirality

end module structure_check_mod
